use minifb::{Scale, ScaleMode, Window, WindowOptions};
use serde::{Deserialize, Serialize};
use termion::{
    color::{self, Bg, Color, Fg},
    cursor::Goto,
    screen::IntoAlternateScreen,
};

use crate::{
    cli::Config,
    cpu::CpuController,
    keyboard::Keyboard,
    memory::{MemHandle, MemoryReader},
    types::{Byte, Word},
};
use std::{
    fs::File,
    io::{self, Read, Write},
    path::Path,
    process::exit,
    thread::sleep,
    time::{Duration, Instant},
};

const FG_COLOR: u32 = 0xFFCC00;
const BG_COLOR: u32 = 0x110500;
const WIDTH: usize = 512;
const HEIGHT: usize = 380;

pub fn get_char_bin<P>(char_rom: Option<P>) -> [u8; 0x8000]
where
    P: AsRef<Path>,
{
    match char_rom {
        Some(path) => {
            let mut file = File::open(path).unwrap();
            let mut bin = vec![0; 0x8000];
            file.read_exact(&mut bin).unwrap();
            // println!("reading char rom");
            bin.try_into().unwrap()
        }
        None => *include_bytes!("./roms/cozette.rom"),
    }
}

trait Renderer {
    fn render(&mut self) {}
}

#[derive(Serialize, Deserialize, Debug, Default)]
pub enum ScreenType {
    Window,
    #[default]
    Terminal,
}

pub struct WindowRenderer {
    char_rom: [u8; 0x8000],
    window: Window,
    memory: MemHandle,
}

impl WindowRenderer {
    pub fn new<P>(memory: MemHandle, char_rom: Option<P>, window: Window) -> Self
    where
        P: AsRef<Path>,
    {
        let char_rom = get_char_bin(char_rom);
        Self {
            memory,
            window,
            char_rom,
        }
    }
}

impl Renderer for WindowRenderer {
    fn render(&mut self) {
        // the rest of this function is arcane wizardry
        // based on the specifics of george's weird
        // display and characters... don't fuck around w it
        let mut i = 0;
        let mut buffer = [0; 512 * 380];
        for char_row in 0..29 {
            for char_col in 0..64 {
                let ascii = self.read(0x6000 + i);
                i += 1;
                for row in 0..13 {
                    let byte = self.char_rom[ascii as usize + (row * 0x100)];
                    for bit_index in (0..8).rev() {
                        let buffer_index =
                            ((char_row) * 13 + (row)) * 512 + (char_col * 8 + bit_index);
                        if (byte << bit_index) & 0x80 == 0x80 {
                            buffer[buffer_index] = FG_COLOR;
                        } else {
                            buffer[buffer_index] = BG_COLOR;
                        }
                    }
                }
            }
        }
        self.window
            .update_with_buffer(&buffer, WIDTH, HEIGHT)
            .unwrap();
    }
}

impl MemoryReader for WindowRenderer {
    fn read(&self, address: Word) -> Byte {
        self.memory.read(address)
    }
}

pub struct TerminalRenderer {
    memory: MemHandle,
}

impl TerminalRenderer {
    pub fn new(memory: MemHandle) -> Self {
        Self { memory }
    }
}

impl MemoryReader for TerminalRenderer {
    fn read(&self, address: Word) -> Byte {
        self.memory.read(address)
    }
}

const ASCII_LOOPUP: [&str; 256] = [
    " ", "░", "▒", "▓", "♡", "♥", "⭐", "✭", "", "✦", "✨", "♀", "♂", "⚢", "⚣", "⚥", "♩", "♪",
    "♫", "♬", "ﱝ", "", "", "", "奄", "奔", "婢", "ﱜ", "ﱛ", "", "", "", " ", "!", "\"", "#",
    "$", "%", "&", "\'", "(", ")", "*", "+", ",", "-", ".", "/", "0", "1", "2", "3", "4", "5", "6",
    "7", "8", "9", ":", ";", "<", "=", ">", "?", "@", "A", "B", "C", "D", "E", "F", "G", "H", "I",
    "J", "K", "L", "M", "N", "O", "P", "Q", "R", "S", "T", "U", "V", "W", "X", "Y", "Z", "[", "\\",
    "]", "^", "_", "`", "a", "b", "c", "d", "e", "f", "g", "h", "i", "j", "k", "l", "m", "n", "o",
    "p", "q", "r", "s", "t", "u", "v", "w", "x", "y", "z", "{", "|", "}", "~", "─", "│", "┌", "└",
    "├", "┤", "┬", "┴", "┼", "╭", "╮", "╯", "╰", "╱", "╲", "╳", "═", "║", "╔", "╗", "╚", "╝", "╠",
    "╣", "╦", "╩", "╬", "", "", "", "", "", "", "", "", "", "", "", "", "", "▁", "▂",
    "▃", "▄", "▅", "▆", "▇", "█", "▉", "▊", "▋", "▌", "▍", "▎", "▏", "ʕ", "·", "ᴥ", "ʔ", "▖", "▗",
    "▘", "▙", "▚", "▛", "▜", "▝", "▞", "▟", "←", "↑", "→", "↓", "⭠", "⭡", "⭢", "⭣", "⮀", "⮁", "⮂",
    "⮃", "", "", "⏳", "", "", "", "", "", "", "", "", "", "", "", "🔮", "", "",
    "◇", "◈", "🌑", "🌒", "🌓", "🌔", "🌕", "🌖", "🌗", "🌘", "", "✔", "✘", "◆", "", "", "",
    "", "", "", "", "🎁", "", "", "", "", "⚐", "⚑", "", "", "",
];

impl Renderer for TerminalRenderer {
    fn render(&mut self) {
        let mut stdout = io::stdout();
        let mut i = 0;
        for char_row in 0..29 {
            for char_col in 0..64 {
                let ascii = self.read(0x6000 + i);
                i += 1;
                let char = ASCII_LOOPUP[ascii as usize];
                let _ = write!(
                    // FG_COLOR = 0xFFCC00
                    // BG_COLOR = 0x110500
                    stdout,
                    "{}{}{}{char}",
                    Goto(char_col + 1, char_row + 1),
                    Fg(color::Rgb(0xFF, 0xCC, 0x00)),
                    Bg(color::Rgb(0x11, 0x05, 0x00))
                );
            }
        }
    }
}

pub struct Screen {
    renderer: Box<dyn Renderer>,
    controller: CpuController,
}

impl Screen {
    pub fn new(config: &Config, controller: CpuController, memory: MemHandle) -> Self {
        let renderer: Box<dyn Renderer> = match config.screen {
            ScreenType::Window => {
                let mut window = Window::new(
                    "ʕ·ᴥ·ʔ-☆",
                    512,
                    380,
                    WindowOptions {
                        resize: true,
                        borderless: true,
                        title: true,
                        transparency: false,
                        scale: Scale::FitScreen,
                        scale_mode: ScaleMode::AspectRatioStretch,
                        topmost: false,
                        none: true,
                    },
                )
                .unwrap();
                window.set_input_callback(Box::new(Keyboard::new(memory.clone())));
                Box::new(WindowRenderer::new(memory, config.char_rom.clone(), window))
            }
            ScreenType::Terminal => Box::new(TerminalRenderer::new(memory)),
        };

        Self {
            renderer,
            controller,
        }
    }
    pub fn draw(&mut self) {
        // self.controller.irq();
        self.renderer.render();
    }

    // pub fn run(&mut self) {
    //     loop {
    //         // sleep(Duration::from_millis(16));
    //         self.draw();
    //     }
    // }
}
